"""Crypto tests.
"""

from __future__ import division, print_function

from binascii import unhexlify

import pytest

import rarfile

try:
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives.ciphers import (
        Cipher, algorithms, modes,
    )
    def aes_encrypt(key, iv, data):
        ciph = Cipher(algorithms.AES(key), modes.CBC(iv), default_backend())
        enc = ciph.encryptor()
        return enc.update(data)
except ImportError:
    pass


@pytest.mark.skipif(not rarfile._have_crypto, reason="No crypto")
def test_aes128_cbc():
    data = b"0123456789abcdef" * 2
    key = b"\x02" * 16
    iv = b"\x80" * 16

    #encdata = aes_encrypt(key, iv, data)
    encdata = unhexlify("4b0d438b4a1b972bd4ab81cd64674dcce4b0158090fbe616f455354284d53502")

    ctx = rarfile.AES_CBC_Decrypt(key, iv)
    assert ctx.decrypt(encdata) == data


@pytest.mark.skipif(not rarfile._have_crypto, reason="No crypto")
def test_aes256_cbc():
    data = b"0123456789abcdef" * 2
    key = b"\x52" * 32
    iv = b"\x70" * 16

    #encdata = aes_encrypt(key, iv, data)
    encdata = unhexlify("24988f387592e4d95b6eaab013137a221f81b25aa7ecde0ef4f4d7a95f92c250")

    ctx = rarfile.AES_CBC_Decrypt(key, iv)
    assert ctx.decrypt(encdata) == data




"""Test zipfile compat.
"""

import inspect
import sys
import zipfile

import pytest

import rarfile

# dont fail on new python by default
_VERS = [(3, 6), (3, 7), (3, 8)]

_UNSUPPORTED = sys.version_info[:2] not in _VERS

_ignore = set([
    "detach",
    "peek",
    "read1",
    "readinto1",

    "seek",

    # no kwargs
    "readinto",
    "readline",
    "truncate",
    "write",

    # random
    "FileHeader",
    "from_file",
    "testzip",
    "writestr",
])


def load_cls_names(maincls):
    assert inspect.isclass(maincls)
    res = {}
    for cls in inspect.getmro(maincls):
        for name, val in inspect.getmembers(cls):
            if name not in res:
                res[name] = val
    return res


def cleansig(sig):
    res = str(sig).replace(", /", "")
    if "*" in res:
        res = res.split(", *", 1)[0] + ")"
    return res


def compare(rmaincls, zmaincls):
    znames = load_cls_names(zmaincls)
    rnames = load_cls_names(rmaincls)
    for name, zval in znames.items():
        if not inspect.isroutine(zval) or name[0] == "_" or name in _ignore:
            continue
        assert name in rnames, "member not found: \"%s\"" % name

        rval = rnames[name]
        zsig = inspect.signature(zval)
        rsig = inspect.signature(rval)

        zsigstr = cleansig(zsig)
        rsigstr = cleansig(rsig)
        assert zsigstr == rsigstr, "sig differs: %s.%s%s != %s.%s%s" % (
            rmaincls.__name__, name, rsigstr,
            zmaincls.__name__, name, zsigstr)


@pytest.mark.skipif(_UNSUPPORTED, reason="Unsupported for sig checks")
def test_cmp_zipfile():
    compare(rarfile.RarFile, zipfile.ZipFile)


@pytest.mark.skipif(_UNSUPPORTED, reason="Unsupported for sig checks")
def test_cmp_zipextfile():
    compare(rarfile.RarExtFile, zipfile.ZipExtFile)


@pytest.mark.skipif(_UNSUPPORTED, reason="Unsupported for sig checks")
def test_cmp_zipinfo():
    compare(rarfile.RarInfo, zipfile.ZipInfo)




"""Format details.
"""

from datetime import datetime

import pytest

import rarfile


def render_date(dt):
    if isinstance(dt, datetime):
        return dt.isoformat("T")
    elif isinstance(dt, tuple):
        return "%04d-%02d-%02d %02d:%02d:%02d" % dt
    else:
        return dt


def mkitem(**kwargs):
    res = {}
    for k, v in kwargs.items():
        if v is not None:
            res[k] = v
    return res


def dumparc(rf):
    res = []
    for item in rf.infolist():
        info = mkitem(fn=item.filename,
                      file_size=item.file_size,
                      compress_size=item.compress_size,
                      CRC=item.CRC,
                      date_time=render_date(item.date_time),
                      arctime=render_date(item.arctime),
                      mtime=render_date(item.mtime),
                      atime=render_date(item.atime),
                      ctime=render_date(item.ctime),
                      comment=item.comment,
                      extract_version=item.extract_version,
                      compress_type=item.compress_type,
                      mode=item.mode,
                      host_os=item.host_os)
        res.append(info)
    return res


def diffs(a, b):
    if len(a) != len(b):
        return "Different lengths"
    problems = []
    for i, xa in enumerate(a):
        xb = b[i]
        for k in xa:
            if k not in xb:
                problems.append("NewKey(%d,%s)=%r" % (i, k, xa[k]))
        for k in xb:
            if k not in xa:
                problems.append("MissingKey(%d,%s)=%r" % (i, k, xb[k]))
        for k in xa:
            if k in xb and xa[k] != xb[k]:
                problems.append("ErrValue(%d,%s):got=%r/exp=%r" % (i, k, xa[k], xb[k]))
    return "; ".join(problems)


def cmp_struct(a, b):
    assert a == b, diffs(a, b)

#
# test start
#


@pytest.mark.skipif(not rarfile._have_crypto, reason="No crypto")
def test_rar3_header_encryption():
    r = rarfile.RarFile("test/files/rar3-comment-hpsw.rar", "r")
    assert r.needs_password() is True
    assert r.comment is None
    assert r.namelist() == []

    r.setpassword("password")
    assert r.needs_password() is True
    assert r.namelist() == ["file1.txt", "file2.txt"]
    assert r.comment is not None
    assert r.comment == "RARcomment\n"


@pytest.mark.skipif(not rarfile._have_crypto, reason="No crypto")
def test_rar5_header_encryption():
    r = rarfile.RarFile("test/files/rar5-hpsw.rar")
    assert r.needs_password() is True
    assert r.comment is None
    assert r.namelist() == []

    with pytest.raises(rarfile.RarWrongPassword):
        r.setpassword("password222")
    assert r.needs_password() is True

    r.setpassword("password")
    assert r.needs_password() is True
    assert r.namelist() == ["stest1.txt", "stest2.txt"]
    assert r.comment is not None
    assert r.comment == "RAR5 archive - hdr-password\n"
    r.close()


def get_vol_info(extver=20, tz="", hr="11"):
    return [
        mkitem(CRC=1352324940,
               date_time="2016-05-24 %s:42:37%s" % (hr, ""),
               mtime="2016-05-24T%s:42:37%s" % (hr, tz),
               compress_type=48,
               compress_size=205000,
               extract_version=extver,
               file_size=205000,
               mode=33204,
               host_os=3,
               fn="vols/bigfile.txt"),
        mkitem(CRC=3498712966,
               date_time="2016-05-24 %s:42:43%s" % (hr, ""),
               mtime="2016-05-24T%s:42:43%s" % (hr, tz),
               extract_version=extver,
               compress_type=48,
               compress_size=2050,
               file_size=2050,
               mode=33204,
               host_os=3,
               fn="vols/smallfile.txt")]


def test_rar3_vols():
    r = rarfile.RarFile("test/files/rar3-vols.part1.rar")
    assert r.needs_password() is False
    assert r.comment is None
    assert r.strerror() is None
    cmp_struct(dumparc(r), get_vol_info())
    assert r.volumelist() == [
        "test/files/rar3-vols.part1.rar",
        "test/files/rar3-vols.part2.rar",
        "test/files/rar3-vols.part3.rar"]
    with pytest.raises(rarfile.NeedFirstVolume):
        rarfile.RarFile("test/files/rar3-vols.part2.rar")


def test_rar3_oldvols():
    r = rarfile.RarFile("test/files/rar3-old.rar")
    assert r.needs_password() is False
    assert r.comment is None
    assert r.strerror() is None
    cmp_struct(dumparc(r), get_vol_info())
    assert r.volumelist() == [
        "test/files/rar3-old.rar",
        "test/files/rar3-old.r00",
        "test/files/rar3-old.r01"]
    with pytest.raises(rarfile.NeedFirstVolume):
        rarfile.RarFile("test/files/rar3-old.r00")


def test_rar5_vols():
    r = rarfile.RarFile("test/files/rar5-vols.part1.rar")
    assert r.needs_password() is False
    assert r.comment is None
    assert r.strerror() is None
    cmp_struct(dumparc(r), get_vol_info(50, "+00:00", "08"))
    assert r.volumelist() == [
        "test/files/rar5-vols.part1.rar",
        "test/files/rar5-vols.part2.rar",
        "test/files/rar5-vols.part3.rar"]
    with pytest.raises(rarfile.NeedFirstVolume):
        rarfile.RarFile("test/files/rar5-vols.part2.rar")


def expect_ctime(mtime, ctime):
    return [mkitem(
        mtime=mtime,
        date_time=mtime.split(".")[0].replace("T", " "),
        ctime=ctime,
        compress_size=0,
        file_size=0,
        CRC=0,
        fn="afile.txt",
        extract_version=29,
        compress_type=48,
        mode=32,
        host_os=2)]


def test_rar3_ctime0():
    r = rarfile.RarFile("test/files/ctime0.rar")
    cmp_struct(dumparc(r), expect_ctime("2011-05-10T21:28:47.899345100", None))


def test_rar3_ctime1():
    r = rarfile.RarFile("test/files/ctime1.rar")
    cmp_struct(dumparc(r), expect_ctime("2011-05-10T21:28:47.899345100", "2011-05-10T21:28:47"))


def test_rar3_ctime2():
    r = rarfile.RarFile("test/files/ctime2.rar")
    cmp_struct(dumparc(r), expect_ctime("2011-05-10T21:28:47.899345100", "2011-05-10T21:28:47.897843200"))


def test_rar3_ctime3():
    r = rarfile.RarFile("test/files/ctime3.rar")
    cmp_struct(dumparc(r), expect_ctime("2011-05-10T21:28:47.899345100", "2011-05-10T21:28:47.899328"))


def test_rar3_ctime4():
    r = rarfile.RarFile("test/files/ctime4.rar")
    cmp_struct(dumparc(r), expect_ctime("2011-05-10T21:28:47.899345100", "2011-05-10T21:28:47.899345100"))


def test_rar5_ctime5():
    r = rarfile.RarFile("test/files/ctime5.rar")
    inf = r.getinfo("timed.txt")
    assert inf.mtime.isoformat() == "2020-07-30T20:26:59.677675904+00:00"
    assert inf.ctime.isoformat() == "2020-07-30T20:28:19.398867888+00:00"
    assert inf.atime.isoformat() == "2020-07-30T20:27:10.121196721+00:00"


def test_rar5_times():
    r = rarfile.RarFile("test/files/rar5-times.rar")
    cmp_struct(dumparc(r), [mkitem(
        fn="stest1.txt",
        file_size=2048,
        compress_size=55,
        compress_type=rarfile.RAR_M3,
        extract_version=50,
        host_os=rarfile.RAR_OS_UNIX,
        mode=33188,
        date_time="2011-06-12 09:53:33",
        mtime="2011-06-12T09:53:33+00:00",
        atime="2016-05-22T09:12:36+00:00",
        CRC=3317163682
    )])


def test_oldvols():
    assert rarfile._next_oldvol("archive") == "archive.r00"
    assert rarfile._next_oldvol("archive.rar/foo") == "archive.rar/foo.r00"
    assert rarfile._next_oldvol("archive.arr") == "archive.a00"
    assert rarfile._next_oldvol("archive.brar") == "archive.b00"
    assert rarfile._next_oldvol("qq00.part0.rar") == "qq00.part0.r00"
    assert rarfile._next_oldvol("qq00.part0.r00") == "qq00.part0.r01"
    assert rarfile._next_oldvol("qq00.part0.r29") == "qq00.part0.r30"
    assert rarfile._next_oldvol("qq00.part0.r99") == "qq00.part0.s00"


def test_newvols():
    assert rarfile._next_newvol("qq00.part0.rar") == "qq00.part1.rar"
    assert rarfile._next_newvol("qq00.part09.rar") == "qq00.part10.rar"
    assert rarfile._next_newvol("qq00.part99.rar") == "qq00.part100.rar"
    assert rarfile._next_newvol("part20") == "part21.rar"
    assert rarfile._next_newvol("qq00.part3.exe") == "qq00.part4.rar"
    assert rarfile._next_newvol("qq00.part5.sfx") == "qq00.part6.rar"
    assert rarfile._next_newvol("qq00.part6.bin") == "qq00.part7.bin"
    assert rarfile._next_newvol("99") == "100.rar"
    assert rarfile._next_newvol("dir/99.rar") == "dir/100.rar"
    with pytest.raises(rarfile.BadRarName):
        rarfile._next_newvol("qq00.part7.rar/foo")
    with pytest.raises(rarfile.BadRarName):
        rarfile._next_newvol("foo")


def test_newvols_err():
    with pytest.raises(rarfile.BadRarName):
        rarfile._next_newvol("xx.rar")


@pytest.mark.parametrize("fn", ["test/files/rar3-versions.rar", "test/files/rar5-versions.rar"])
def test_versions(fn):
    with rarfile.RarFile(fn) as rf:
        assert rf.namelist() == ["versioned.txt"]




"""API tests.
"""

import io
import os
from pathlib import Path

import pytest

import rarfile

#
# test start
#

def test_not_rar():
    with pytest.raises(rarfile.NotRarFile):
        rarfile.RarFile("rarfile.py", "r")
    with pytest.raises(rarfile.NotRarFile):
        with open("rarfile.py", "rb") as f:
            rarfile.RarFile(f, "r")


def test_bad_arc_mode_w():
    with pytest.raises(NotImplementedError):
        rarfile.RarFile("test/files/rar3-comment-plain.rar", "w")


def test_bad_arc_mode_rb():
    with pytest.raises(NotImplementedError):
        rarfile.RarFile("test/files/rar3-comment-plain.rar", "rb")


def test_bad_errs():
    with pytest.raises(ValueError):
        rarfile.RarFile("test/files/rar3-comment-plain.rar", "r", errors="foo")


def test_errors_param():
    with open("test/files/rar3-comment-plain.rar", "rb") as f:
        data = f.read()
    buf = io.BytesIO(data[:17])
    with rarfile.RarFile(buf, "r", errors="stop") as rf:
        assert rf.namelist() == []
    with pytest.raises(rarfile.BadRarFile):
        rarfile.RarFile(buf, "r", errors="strict")


def test_bad_open_mode_w():
    rf = rarfile.RarFile("test/files/rar3-comment-plain.rar")
    with pytest.raises(NotImplementedError):
        rf.open("qwe", "w")


def test_bad_open_psw():
    rf = rarfile.RarFile("test/files/rar3-comment-psw.rar")
    with pytest.raises(rarfile.PasswordRequired):
        rf.open("file1.txt")


def test_bad_filelike():
    with pytest.raises(ValueError):
        rarfile.is_rarfile(bytearray(10))


def test_open_psw_late_rar3():
    rf = rarfile.RarFile("test/files/rar3-comment-psw.rar")
    d1 = rf.open("file1.txt", "r", "password").read()
    d2 = rf.open("file1.txt", "r", b"password").read()
    assert d1 == d2


def test_open_psw_late_rar5():
    rf = rarfile.RarFile("test/files/rar5-psw.rar")
    rf.open("stest1.txt", "r", "password").read()
    rf.open("stest1.txt", "r", b"password").read()


def test_open_pathlib_path():
    rf = rarfile.RarFile("test/files/rar5-psw.rar")
    rf.open(Path("stest1.txt"), "r", "password").read()


def test_read_psw_late_rar3():
    rf = rarfile.RarFile("test/files/rar3-comment-psw.rar")
    rf.read("file1.txt", "password")
    rf.read("file1.txt", b"password")


def test_read_psw_late_rar5():
    rf = rarfile.RarFile("test/files/rar5-psw.rar")
    rf.read("stest1.txt", "password")
    rf.read("stest1.txt", b"password")


def test_open_psw_late():
    rf = rarfile.RarFile("test/files/rar5-psw.rar")
    with pytest.raises(rarfile.BadRarFile):
        rf.read("stest1.txt", "password222")


def test_create_from_pathlib_path():
    # Make sure we can open both relative and absolute Paths
    rarfile.RarFile(Path("test/files/rar5-psw.rar"))
    rarfile.RarFile(Path("test/files/rar5-psw.rar").resolve())


def test_detection():
    assert rarfile.is_rarfile("test/files/ctime4.rar.exp") is False
    assert rarfile.is_rarfile("test/files/ctime4.rar") is True
    assert rarfile.is_rarfile("test/files/rar5-crc.rar") is True

    assert rarfile.is_rarfile(Path("test/files/rar5-crc.rar")) is True

    assert rarfile.is_rarfile("test/files/_missing_.rar") is False


def test_restore_pos():
    with open('test/files/rar5-crc.rar', 'rb') as fd:
        fd.seek(3)
        assert rarfile.is_rarfile(fd) is True
        assert fd.tell() == 3


def test_getinfo():
    with rarfile.RarFile("test/files/rar5-crc.rar") as rf:
        inf = rf.getinfo("stest1.txt")
        assert isinstance(inf, rarfile.RarInfo)
        assert rf.getinfo(inf) is inf
        with pytest.raises(rarfile.NoRarEntry):
            rf.getinfo("missing.txt")

def test_signature_error():
    with pytest.raises(rarfile.NotRarFile):
        rarfile.RarFile("test/files/ctime4.rar.exp")


def test_signature_error_mem():
    data = io.BytesIO(b"x" * 40)
    with pytest.raises(rarfile.NotRarFile):
        rarfile.RarFile(data)


def test_with():
    with rarfile.RarFile("test/files/rar5-crc.rar") as rf:
        data = rf.read("stest1.txt")
        with rf.open("stest1.txt") as f:
            dst = io.BytesIO()
            while True:
                buf = f.read(7)
                if not buf:
                    break
                dst.write(buf)
            assert dst.getvalue() == data


def test_readline():
    def load_readline(rf, fn):
        with rf.open(fn) as f:
            tr = io.TextIOWrapper(io.BufferedReader(f))
            res = []
            while True:
                ln = tr.readline()
                if not ln:
                    break
                res.append(ln)
        return res

    rf = rarfile.RarFile("test/files/seektest.rar")
    v1 = load_readline(rf, "stest1.txt")
    v2 = load_readline(rf, "stest2.txt")
    assert len(v1) == 512
    assert v1 == v2


def run_parallel(rfile, entry):
    buf1, buf2 = [], []
    rf = rarfile.RarFile(rfile)
    count = 0
    with rf.open(entry) as f1:
        with rf.open(entry) as f2:
            for _ in range(10000):
                res1 = f1.read(10)
                res2 = f2.read(10)
                if res1:
                    buf1.append(res1)
                    count += len(res1)
                if res2:
                    buf2.append(res2)
                    count += len(res2)
                if not res1 and not res2:
                    break

    assert buf1
    assert buf1 == buf2


def test_parallel_file_compressed():
    run_parallel("test/files/seektest.rar", "stest1.txt")


def test_parallel_file_direct():
    run_parallel("test/files/seektest.rar", "stest2.txt")


def test_parallel_fd_compressed():
    with open("test/files/seektest.rar", "rb") as f:
        memfile = io.BytesIO(f.read())
    run_parallel(memfile, "stest1.txt")


def test_parallel_fd_direct():
    with open("test/files/seektest.rar", "rb") as f:
        memfile = io.BytesIO(f.read())
    run_parallel(memfile, "stest2.txt")


def test_printdir(capsys):
    rf = rarfile.RarFile("test/files/seektest.rar")
    rf.printdir()
    res = capsys.readouterr()
    assert res.out == "stest1.txt\nstest2.txt\n"


def test_testrar():
    rf = rarfile.RarFile("test/files/seektest.rar")
    rf.testrar()


def test_iter():
    rf = rarfile.RarFile("test/files/seektest.rar")
    n1 = rf.namelist()
    n2 = [m.filename for m in rf]
    assert n1 == n2


def test_testrar_mem():
    with open("test/files/seektest.rar", "rb") as f:
        arc = f.read()
    rf = rarfile.RarFile(io.BytesIO(arc))
    rf.testrar()


def test_extract(tmp_path):
    ex1 = tmp_path / "extract1"
    ex2 = tmp_path / "extract2"
    ex3 = tmp_path / "extract3"
    os.makedirs(str(ex1))
    os.makedirs(str(ex2))
    os.makedirs(str(ex3))
    rf = rarfile.RarFile("test/files/seektest.rar")

    rf.extractall(str(ex1))
    assert os.path.isfile(str(ex1 / "stest1.txt")) is True
    assert os.path.isfile(str(ex1 / "stest2.txt")) is True

    rf.extract("stest1.txt", str(ex2))
    assert os.path.isfile(str(ex2 / "stest1.txt")) is True
    assert os.path.isfile(str(ex2 / "stest2.txt")) is False

    inf = rf.getinfo("stest2.txt")
    rf.extract(inf, str(ex3))
    assert os.path.isfile(str(ex3 / "stest1.txt")) is False
    assert os.path.isfile(str(ex3 / "stest2.txt")) is True

    rf.extractall(str(ex2), ["stest1.txt"])
    assert os.path.isfile(str(ex2 / "stest1.txt")) is True

    rf.extractall(str(ex3), [rf.getinfo("stest2.txt")])
    assert os.path.isfile(str(ex3 / "stest2.txt")) is True

    ex4 = tmp_path / "extract4"
    os.makedirs(str(ex4))
    rf.extractall(ex4)
    assert os.path.isfile(str(ex4 / "stest1.txt")) is True
    assert os.path.isfile(str(ex4 / "stest2.txt")) is True


def test_extract_mem(tmp_path):
    ex1 = tmp_path / "extract11"
    ex2 = tmp_path / "extract22"
    ex3 = tmp_path / "extract33"
    os.makedirs(str(ex1))
    os.makedirs(str(ex2))
    os.makedirs(str(ex3))

    with open("test/files/seektest.rar", "rb") as f:
        arc = f.read()
    rf = rarfile.RarFile(io.BytesIO(arc))

    rf.extractall(str(ex1))
    assert os.path.isfile(str(ex1 / "stest1.txt")) is True
    assert os.path.isfile(str(ex1 / "stest2.txt")) is True

    rf.extract("stest1.txt", str(ex2))
    assert os.path.isfile(str(ex2 / "stest1.txt")) is True
    assert os.path.isfile(str(ex2 / "stest2.txt")) is False

    inf = rf.getinfo("stest2.txt")
    rf.extract(inf, str(ex3))
    assert os.path.isfile(str(ex3 / "stest1.txt")) is False
    assert os.path.isfile(str(ex3 / "stest2.txt")) is True


def get_rftype(h):
    assert h.is_dir() == h.isdir()
    return "".join([
        h.is_file() and "F" or "-",
        h.is_dir() and "D" or "-",
        h.is_symlink() and "L" or "-",
    ])


def test_infocb():
    infos = []

    def info_cb(info):
        infos.append((info.type, info.needs_password(), get_rftype(info), info._must_disable_hack()))

    rf = rarfile.RarFile("test/files/seektest.rar", info_callback=info_cb)
    assert infos == [
        (rarfile.RAR_BLOCK_MAIN, False, "---", False),
        (rarfile.RAR_BLOCK_FILE, False, "F--", False),
        (rarfile.RAR_BLOCK_FILE, False, "F--", False),
        (rarfile.RAR_BLOCK_ENDARC, False, "---", False)]
    rf.close()

    infos = []
    rf = rarfile.RarFile("test/files/rar5-solid-qo.rar", info_callback=info_cb)
    assert infos == [
        (rarfile.RAR_BLOCK_MAIN, False, "---", True),
        (rarfile.RAR_BLOCK_FILE, False, "F--", False),
        (rarfile.RAR_BLOCK_FILE, False, "F--", True),
        (rarfile.RAR_BLOCK_FILE, False, "F--", True),
        (rarfile.RAR_BLOCK_FILE, False, "F--", True),
        (rarfile.RAR_BLOCK_SUB, False, "---", False),
        (rarfile.RAR_BLOCK_ENDARC, False, "---", False)]
    rf.close()


# pylint: disable=singleton-comparison
def test_rarextfile():
    with rarfile.RarFile("test/files/seektest.rar") as rf:
        for fn in ("stest1.txt", "stest2.txt"):
            with rf.open(fn) as f:
                assert f.tell() == 0
                assert f.writable() == False
                assert f.seekable() == True
                assert f.readable() == True
                assert f.readall() == rf.read(fn)


def test_is_rarfile():
    with rarfile.RarFile("test/files/seektest.rar") as rf:
        for fn in ("stest1.txt", "stest2.txt"):
            with rf.open(fn) as f:
                assert f.tell() == 0
                assert f.writable() == False
                assert f.seekable() == True
                assert f.readable() == True
                assert f.readall() == rf.read(fn)


def test_part_only():
    info_list = []
    def info_cb(info):
        info_list.append(info)

    with pytest.raises(rarfile.NeedFirstVolume):
        with rarfile.RarFile("test/files/rar3-vols.part2.rar") as rf:
            pass
    with rarfile.RarFile("test/files/rar3-vols.part2.rar", part_only=True, info_callback=info_cb) as rf:
        assert len(info_list) == 3

    with pytest.raises(rarfile.NeedFirstVolume):
        with rarfile.RarFile("test/files/rar5-vols.part2.rar") as rf:
            pass
    info_list = []
    with rarfile.RarFile("test/files/rar5-vols.part2.rar", part_only=True, info_callback=info_cb) as rf:
        assert len(info_list) == 5


def test_volume_info():
    info_list = []
    def info_cb(info):
        info_list.append(info)
    with rarfile.RarFile("test/files/rar3-vols.part1.rar", info_callback=info_cb) as rf:
        assert len(info_list) == 10
    info_list = []
    with rarfile.RarFile("test/files/rar5-vols.part1.rar", info_callback=info_cb) as rf:
        assert len(info_list) == 16


def test_is_solid():
    with rarfile.RarFile("test/files/rar3-comment-plain.rar") as rf:
        assert not rf.is_solid()
    with rarfile.RarFile("test/files/rar3-solid.rar") as rf:
        assert rf.is_solid()
    with rarfile.RarFile("test/files/rar5-crc.rar") as rf:
        assert not rf.is_solid()
    with rarfile.RarFile("test/files/rar5-solid.rar") as rf:
        assert rf.is_solid()




"""test corrupt file parsing.
"""

import glob
import io

import rarfile


def try_read(tmpfn):
    if not rarfile.is_rarfile(tmpfn):
        return
    rarfile.RarFile(tmpfn, errors="stop")
    try:
        rf = rarfile.RarFile(tmpfn, errors="strict")
        if rf.needs_password():
            rf.setpassword("password")
    except rarfile.Error:
        return
    for fn in rf.namelist():
        try:
            rf.read(fn)
        except rarfile.Error:
            pass


def process_rar(rarfn, quick=False):
    with open(rarfn, "rb") as f:
        data = f.read()
    for n in range(len(data)):
        bad = io.BytesIO(data[:n])
        if not rarfile.is_rarfile(bad):
            return
        try_read(bad)

    crap = b"\x00\xff\x01\x80\x7f"
    if quick:
        crap = b"\xff"
    for n in range(1, len(data)):
        for i in range(len(crap)):
            c = crap[i:i + 1]
            bad = data[:n - 1] + c + data[n:]
            try_read(io.BytesIO(bad))


def test_corrupt_quick_rar3():
    process_rar("test/files/rar3-comment-plain.rar", True)


def test_corrupt_quick_rar5():
    process_rar("test/files/rar5-times.rar", True)


def test_corrupt_all():
    test_rar_list = glob.glob("test/files/*.rar")
    test_rar_list = []
    for rar in test_rar_list:
        process_rar(rar)


if __name__ == "__main__":
    test_corrupt_quick_rar5()




#! /usr/bin/env python3

"""Dump archive contents, test extraction."""

import binascii
import getopt
import io
import sys
from datetime import datetime

import rarfile as rf

usage = """
dumprar [switches] [ARC1 ARC2 ...] [@ARCLIST]
switches:
  @file      read archive names from file
  -pPWD      set password
  -Ccharset  set fallback charset
  -v         increase verbosity
  -t         attempt to read all files
  -x         write read files out
  -c         show archive comment
  -h         show usage
  -bTOOL     set backend tool (unrar, unar, bsdtar, 7z, 7zz)
  --         stop switch parsing
""".strip()

os_list = ["DOS", "OS2", "WIN", "UNIX", "MACOS", "BEOS"]

block_strs = ["MARK", "MAIN", "FILE", "OLD_COMMENT", "OLD_EXTRA",
              "OLD_SUB", "OLD_RECOVERY", "OLD_AUTH", "SUB", "ENDARC"]

r5_block_types = {
    rf.RAR5_BLOCK_MAIN: "R5_MAIN",
    rf.RAR5_BLOCK_FILE: "R5_FILE",
    rf.RAR5_BLOCK_SERVICE: "R5_SVC",
    rf.RAR5_BLOCK_ENCRYPTION: "R5_ENC",
    rf.RAR5_BLOCK_ENDARC: "R5_ENDARC",
}


def rar3_type(btype):
    """RAR3 type code as string."""
    if btype < rf.RAR_BLOCK_MARK or btype > rf.RAR_BLOCK_ENDARC:
        return "*UNKNOWN*"
    return block_strs[btype - rf.RAR_BLOCK_MARK]


def rar5_type(btype):
    """RAR5 type code as string."""
    return r5_block_types.get(btype, "*UNKNOWN*")


main_bits = (
    (rf.RAR_MAIN_VOLUME, "VOL"),
    (rf.RAR_MAIN_COMMENT, "COMMENT"),
    (rf.RAR_MAIN_LOCK, "LOCK"),
    (rf.RAR_MAIN_SOLID, "SOLID"),
    (rf.RAR_MAIN_NEWNUMBERING, "NEWNR"),
    (rf.RAR_MAIN_AUTH, "AUTH"),
    (rf.RAR_MAIN_RECOVERY, "RECOVERY"),
    (rf.RAR_MAIN_PASSWORD, "PASSWORD"),
    (rf.RAR_MAIN_FIRSTVOLUME, "FIRSTVOL"),
    (rf.RAR_SKIP_IF_UNKNOWN, "SKIP"),
    (rf.RAR_LONG_BLOCK, "LONG"),
)

endarc_bits = (
    (rf.RAR_ENDARC_NEXT_VOLUME, "NEXTVOL"),
    (rf.RAR_ENDARC_DATACRC, "DATACRC"),
    (rf.RAR_ENDARC_REVSPACE, "REVSPACE"),
    (rf.RAR_ENDARC_VOLNR, "VOLNR"),
    (rf.RAR_SKIP_IF_UNKNOWN, "SKIP"),
    (rf.RAR_LONG_BLOCK, "LONG"),
)

file_bits = (
    (rf.RAR_FILE_SPLIT_BEFORE, "SPLIT_BEFORE"),
    (rf.RAR_FILE_SPLIT_AFTER, "SPLIT_AFTER"),
    (rf.RAR_FILE_PASSWORD, "PASSWORD"),
    (rf.RAR_FILE_COMMENT, "COMMENT"),
    (rf.RAR_FILE_SOLID, "SOLID"),
    (rf.RAR_FILE_LARGE, "LARGE"),
    (rf.RAR_FILE_UNICODE, "UNICODE"),
    (rf.RAR_FILE_SALT, "SALT"),
    (rf.RAR_FILE_VERSION, "VERSION"),
    (rf.RAR_FILE_EXTTIME, "EXTTIME"),
    (rf.RAR_FILE_EXTFLAGS, "EXTFLAGS"),
    (rf.RAR_SKIP_IF_UNKNOWN, "SKIP"),
    (rf.RAR_LONG_BLOCK, "LONG"),
)

generic_bits = (
    (rf.RAR_SKIP_IF_UNKNOWN, "SKIP"),
    (rf.RAR_LONG_BLOCK, "LONG"),
)

file_parms = ("D64", "D128", "D256", "D512",
              "D1024", "D2048", "D4096", "DIR")

r5_block_flags = (
    (rf.RAR5_BLOCK_FLAG_EXTRA_DATA, "EXTRA"),
    (rf.RAR5_BLOCK_FLAG_DATA_AREA, "DATA"),
    (rf.RAR5_BLOCK_FLAG_SKIP_IF_UNKNOWN, "SKIP"),
    (rf.RAR5_BLOCK_FLAG_SPLIT_BEFORE, "SPLIT_BEFORE"),
    (rf.RAR5_BLOCK_FLAG_SPLIT_AFTER, "SPLIT_AFTER"),
    (rf.RAR5_BLOCK_FLAG_DEPENDS_PREV, "DEPENDS"),
    (rf.RAR5_BLOCK_FLAG_KEEP_WITH_PARENT, "KEEP"),
)

r5_main_flags = (
    (rf.RAR5_MAIN_FLAG_ISVOL, "ISVOL"),
    (rf.RAR5_MAIN_FLAG_HAS_VOLNR, "VOLNR"),
    (rf.RAR5_MAIN_FLAG_SOLID, "SOLID"),
    (rf.RAR5_MAIN_FLAG_RECOVERY, "RECOVERY"),
    (rf.RAR5_MAIN_FLAG_LOCKED, "LOCKED"),
)

r5_file_flags = (
    (rf.RAR5_FILE_FLAG_ISDIR, "DIR"),
    (rf.RAR5_FILE_FLAG_HAS_MTIME, "MTIME"),
    (rf.RAR5_FILE_FLAG_HAS_CRC32, "CRC32"),
    (rf.RAR5_FILE_FLAG_UNKNOWN_SIZE, "NOSIZE"),
)

r5_enc_flags = (
    (rf.RAR5_ENC_FLAG_HAS_CHECKVAL, "CHECKVAL"),
)

r5_endarc_flags = (
    (rf.RAR5_ENDARC_FLAG_NEXT_VOL, "NEXTVOL"),
)

r5_file_enc_flags = (
    (rf.RAR5_XENC_CHECKVAL, "CHECKVAL"),
    (rf.RAR5_XENC_TWEAKED, "TWEAKED"),
)

r5_file_redir_types = {
    rf.RAR5_XREDIR_UNIX_SYMLINK: "UNIX_SYMLINK",
    rf.RAR5_XREDIR_WINDOWS_SYMLINK: "WINDOWS_SYMLINK",
    rf.RAR5_XREDIR_WINDOWS_JUNCTION: "WINDOWS_JUNCTION",
    rf.RAR5_XREDIR_HARD_LINK: "HARD_LINK",
    rf.RAR5_XREDIR_FILE_COPY: "FILE_COPY",
}

r5_file_redir_flags = (
    (rf.RAR5_XREDIR_ISDIR, "DIR"),
)

r2_subblock_types = {
    rf.RAR_OLD_SUB_OS2: "OS2",
    rf.RAR_OLD_SUB_UNIX: "UNIX",
    rf.RAR_OLD_SUB_MAC: "MAC",
    rf.RAR_OLD_SUB_BEOS: "BEOS",
    rf.RAR_OLD_SUB_NT: "NT",
    rf.RAR_OLD_SUB_STREAM: "STREAM",
}

dos_mode_bits = (
    (0x01, "READONLY"),
    (0x02, "HIDDEN"),
    (0x04, "SYSTEM"),
    (0x08, "VOLUME_ID"),
    (0x10, "DIRECTORY"),
    (0x20, "ARCHIVE"),
    (0x40, "DEVICE"),
    (0x80, "NORMAL"),
    (0x0100, "TEMPORARY"),
    (0x0200, "SPARSE_FILE"),
    (0x0400, "REPARSE_POINT"),
    (0x0800, "COMPRESSED"),
    (0x1000, "OFFLINE"),
    (0x2000, "NOT_CONTENT_INDEXED"),
    (0x4000, "ENCRYPTED"),
    (0x8000, "INTEGRITY_STREAM"),
    (0x00010000, "VIRTUAL"),
    (0x00020000, "NO_SCRUB_DATA"),
    (0x00040000, "RECALL_ON_OPEN"),
    (0x00080000, "PINNED"),
    (0x00100000, "UNPINNED"),
    (0x00400000, "RECALL_ON_DATA_ACCESS"),
    (0x20000000, "STRICTLY_SEQUENTIAL"),
)


def xprint(m, *args):
    """Print string to stdout.
    """
    if args:
        m = m % args
    print(m)


def tohex(data):
    """Return hex string."""
    return binascii.hexlify(data).decode("ascii")


def render_flags(flags, bit_list):
    """Show bit names.
    """
    res = []
    known = 0
    for bit in bit_list:
        known = known | bit[0]
        if flags & bit[0]:
            res.append(bit[1])
    unknown = flags & ~known
    n = 0
    while unknown:
        if unknown & 1:
            res.append("UNK_%04x" % (1 << n))
        unknown = unknown >> 1
        n += 1

    if not res:
        return "-"

    return ",".join(res)


def get_file_flags(flags):
    """Show flag names and handle dict size.
    """
    res = render_flags(flags & ~rf.RAR_FILE_DICTMASK, file_bits)

    xf = (flags & rf.RAR_FILE_DICTMASK) >> 5
    res += "," + file_parms[xf]
    return res


def fmt_time(t):
    """Format time.
    """
    if t is None:
        return "(-)"
    if isinstance(t, datetime):
        return t.isoformat("T")
    return "%04d-%02d-%02d %02d:%02d:%02d" % t


def show_item(h):
    """Show any RAR3/5 record.
    """
    if isinstance(h, rf.Rar3Info):
        show_item_v3(h)
    elif isinstance(h, rf.Rar5Info):
        show_item_v5(h)
    else:
        xprint("Unknown info record")


def show_rftype(h):
    return "".join([
        h.is_file() and "F" or "-",
        h.is_dir() and "D" or "-",
        h.is_symlink() and "L" or "-",
    ])


def modex3(v):
    return [v & 4 and "r" or "-", v & 2 and "w" or "-", v & 1 and "x" or "-"]


def unix_mode(mode):
    perms = modex3(mode >> 6) + modex3(mode >> 3) + modex3(mode)
    if mode & 0x0800:
        perms[2] = perms[2] == "x" and "s" or "S"
    if mode & 0x0400:
        perms[5] = perms[5] == "x" and "s" or "S"
    if mode & 0x0200:
        perms[8] = perms[8] == "x" and "t" or "-"
    rest = mode & 0xF000
    if rest == 0x4000:
        perms.insert(0, "d")
    elif rest == 0xA000:
        perms.insert(0, "l")
    elif rest == 0x8000:
        # common
        perms.insert(0, "-")
    elif rest == 0:
        perms.insert(0, "-")
    else:
        perms.insert(0, "?")
        perms.append("(0x%04x)" % rest)
    return "".join(perms)


def show_mode(h):
    if h.host_os in (rf.RAR_OS_UNIX, rf.RAR_OS_BEOS):
        s_mode = unix_mode(h.mode)
    elif h.host_os in (rf.RAR_OS_MSDOS, rf.RAR_OS_WIN32, rf.RAR_OS_OS2):
        s_mode = render_flags(h.mode, dos_mode_bits)
    else:
        s_mode = "0x%x" % h.mode
    return s_mode


def show_item_v3(h):
    """Show any RAR3 record.
    """
    st = rar3_type(h.type)
    xprint("%s: hdrlen=%d datlen=%d is=%s",
           st, h.header_size, h.add_size, show_rftype(h))
    if h.type in (rf.RAR_BLOCK_FILE, rf.RAR_BLOCK_SUB):
        s_mode = show_mode(h)
        xprint("  flags=0x%04x:%s", h.flags, get_file_flags(h.flags))
        if h.host_os >= 0 and h.host_os < len(os_list):
            s_os = os_list[h.host_os]
        else:
            s_os = "?"
        if h.flags & rf.RAR_FILE_UNICODE:
            s_namecmp = " namecmp=%d/%d" % (len(h.orig_filename), h._name_size)
        else:
            s_namecmp = ""
        xprint("  os=%d:%s ver=%d mode=%s meth=%c cmp=%d dec=%d vol=%d%s",
               h.host_os, s_os,
               h.extract_version, s_mode, h.compress_type,
               h.compress_size, h.file_size, h.volume, s_namecmp)
        ucrc = (h.CRC + (1 << 32)) & ((1 << 32) - 1)
        xprint("  crc=0x%08x (%d) date_time=%s", ucrc, h.CRC, fmt_time(h.date_time))
        xprint("  name=%s", h.filename)
        if h.mtime:
            xprint("  mtime=%s", fmt_time(h.mtime))
        if h.ctime:
            xprint("  ctime=%s", fmt_time(h.ctime))
        if h.atime:
            xprint("  atime=%s", fmt_time(h.atime))
        if h.arctime:
            xprint("  arctime=%s", fmt_time(h.arctime))
    elif h.type == rf.RAR_BLOCK_MAIN:
        xprint("  flags=0x%04x:%s", h.flags, render_flags(h.flags, main_bits))
    elif h.type == rf.RAR_BLOCK_ENDARC:
        xprint("  flags=0x%04x:%s", h.flags, render_flags(h.flags, endarc_bits))
        if h.flags & rf.RAR_ENDARC_DATACRC:
            xprint("  datacrc=0x%08x", h.endarc_datacrc)
        if h.flags & rf.RAR_ENDARC_DATACRC:
            xprint("  volnr=%d", h.endarc_volnr)
    elif h.type == rf.RAR_BLOCK_MARK:
        xprint("  flags=0x%04x:", h.flags)
    elif h.type == rf.RAR_BLOCK_OLD_SUB:
        xprint("  flags=0x%04x:%s", h.flags, render_flags(h.flags, generic_bits))
        xprint("  sub_type=0x%04x:%s", h.old_sub_type,
               r2_subblock_types.get(h.old_sub_type, '*UNKNOWN*'))
    else:
        xprint("  flags=0x%04x:%s", h.flags, render_flags(h.flags, generic_bits))

    if h.comment is not None:
        cm = repr(h.comment)
        if cm[0] == "u":
            cm = cm[1:]
        xprint("  comment=%s", cm)


def show_item_v5(h):
    """Show any RAR5 record.
    """
    st = rar5_type(h.block_type)
    xprint("%s: hdrlen=%d datlen=%d hdr_extra=%d is=%s", st, h.header_size,
           h.compress_size, h.block_extra_size, show_rftype(h))
    xprint("  block_flags=0x%04x:%s", h.block_flags, render_flags(h.block_flags, r5_block_flags))
    if h.block_type in (rf.RAR5_BLOCK_FILE, rf.RAR5_BLOCK_SERVICE):
        xprint("  name=%s", h.filename)
        s_mode = show_mode(h)
        if h.file_host_os == rf.RAR5_OS_UNIX:
            s_os = "UNIX"
        else:
            s_os = "WINDOWS"
        xprint("  file_flags=0x%04x:%s", h.file_flags, render_flags(h.file_flags, r5_file_flags))

        cmp_flags = h.file_compress_flags
        xprint("  cmp_algo=%d cmp_meth=%d dict=%d solid=%r",
               cmp_flags & 0x3f,
               (cmp_flags >> 7) & 0x07,
               cmp_flags >> 10,
               cmp_flags & rf.RAR5_COMPR_SOLID > 0)
        xprint("  os=%d:%s mode=%s cmp=%r dec=%r vol=%r",
               h.file_host_os, s_os, s_mode,
               h.compress_size, h.file_size, h.volume)
        if h.CRC is not None:
            xprint("  crc=0x%08x (%d)", h.CRC, h.CRC)
        if h.blake2sp_hash is not None:
            xprint("  blake2sp=%s", tohex(h.blake2sp_hash))
        if h.date_time is not None:
            xprint("  date_time=%s", fmt_time(h.date_time))
        if h.mtime:
            xprint("  mtime=%s", fmt_time(h.mtime))
        if h.ctime:
            xprint("  ctime=%s", fmt_time(h.ctime))
        if h.atime:
            xprint("  atime=%s", fmt_time(h.atime))
        if h.arctime:
            xprint("  arctime=%s", fmt_time(h.arctime))
        if h.flags & rf.RAR_FILE_PASSWORD:
            enc_algo, enc_flags, kdf_count, salt, iv, checkval = h.file_encryption
            algo_name = "AES256" if enc_algo == rf.RAR5_XENC_CIPHER_AES256 else "UnknownAlgo"
            xprint("  algo=%d:%s enc_flags=%04x:%s kdf_lg=%d kdf_count=%d salt=%s iv=%s checkval=%s",
                   enc_algo, algo_name, enc_flags, render_flags(enc_flags, r5_file_enc_flags),
                   kdf_count, 1 << kdf_count, tohex(salt), tohex(iv),
                   checkval and tohex(checkval) or "-")
        if h.file_redir:
            redir_type, redir_flags, redir_name = h.file_redir
            xprint("  redir: type=%s flags=%d:%s destination=%s",
                   r5_file_redir_types.get(redir_type, "Unknown"),
                   redir_flags, render_flags(redir_flags, r5_file_redir_flags),
                   redir_name)
        if h.file_owner:
            uname, gname, uid, gid = h.file_owner
            xprint("  owner: name=%r group=%r uid=%r gid=%r",
                   uname, gname, uid, gid)
        if h.file_version:
            flags, version = h.file_version
            xprint("  version: flags=%r version=%r", flags, version)
    elif h.block_type == rf.RAR5_BLOCK_MAIN:
        xprint("  flags=0x%04x:%s", h.flags, render_flags(h.main_flags, r5_main_flags))
    elif h.block_type == rf.RAR5_BLOCK_ENDARC:
        xprint("  flags=0x%04x:%s", h.flags, render_flags(h.endarc_flags, r5_endarc_flags))
    elif h.block_type == rf.RAR5_BLOCK_ENCRYPTION:
        algo_name = "AES256" if h.encryption_algo == rf.RAR5_XENC_CIPHER_AES256 else "UnknownAlgo"
        xprint("  algo=%d:%s flags=0x%04x:%s", h.encryption_algo, algo_name, h.flags,
               render_flags(h.encryption_flags, r5_enc_flags))
        xprint("  kdf_lg=%d kdf_count=%d", h.encryption_kdf_count, 1 << h.encryption_kdf_count)
        xprint("  salt=%s", tohex(h.encryption_salt))
    else:
        xprint("  - missing info -")

    if h.comment is not None:
        cm = repr(h.comment)
        if cm[0] == "u":
            cm = cm[1:]
        xprint("  comment=%s", cm)


cf_show_comment = 0
cf_verbose = 0
cf_charset = None
cf_extract = 0
cf_test_read = 0
cf_test_unrar = 0
cf_test_memory = 0


def check_crc(f, inf, desc):
    """Compare result crc to expected value.
    """
    exp = inf._md_expect
    if exp is None:
        return
    ucrc = f._md_context.digest()
    if ucrc != exp:
        print("crc error - %s - exp=%r got=%r" % (desc, exp, ucrc))


def test_read_long(r, inf):
    """Test read and readinto.
    """
    md_class = inf._md_class or rf.NoHashContext
    bctx = md_class()
    inf_orig = r.getinfo_orig(inf.filename)
    f = r.open(inf.filename)
    total = 0
    while 1:
        data = f.read(8192)
        if not data:
            break
        bctx.update(data)
        total += len(data)
    if total != inf.file_size:
        xprint("\n *** %s has corrupt file: %s ***", r.rarfile, inf.filename)
        xprint(" *** short read: got=%d, need=%d ***\n", total, inf.file_size)
    check_crc(f, inf_orig, "read")
    bhash = bctx.hexdigest()
    if cf_verbose > 1:
        if f._md_context.digest() == inf_orig._md_expect:
            #xprint("  checkhash: %r", bhash)
            pass
        else:
            xprint("  checkhash: %r  got=%r exp=%r cls=%r\n",
                   bhash, f._md_context.digest(), inf._md_expect, inf._md_class)

    # test .seek() & .readinto()
    if cf_test_read > 1:
        f.seek(0, 0)

        total = 0
        buf = bytearray(1024)
        while 1:
            res = f.readinto(buf)
            if not res:
                break
            total += res
        if inf.file_size != total:
            xprint(" *** readinto failed: got=%d, need=%d ***\n", total, inf.file_size)
        #check_crc(f, inf, "readinto")
    f.close()


def test_read(r, inf):
    """Test file read."""
    test_read_long(r, inf)


def test_real(fn, pwd):
    """Actual archive processing.
    """
    xprint("Archive: %s", fn)

    cb = None
    if cf_verbose > 1:
        cb = show_item

    rfarg = fn
    if cf_test_memory:
        rfarg = io.BytesIO(open(fn, "rb").read())

    # check if rar
    if not rf.is_rarfile(rfarg):
        xprint(" --- %s is not a RAR file ---", fn)
        return

    # open
    r = rf.RarFile(rfarg, charset=cf_charset, info_callback=cb)
    # set password
    if r.needs_password():
        if pwd:
            r.setpassword(pwd)
        else:
            xprint(" --- %s requires password ---", fn)
            return

    # show comment
    if cf_show_comment and r.comment:
        for ln in r.comment.split("\n"):
            xprint("    %s", ln)
    elif cf_verbose > 0 and r.comment:
        cm = repr(r.comment)
        if cm[0] == "u":
            cm = cm[1:]
        xprint("  comment=%s", cm)

    # process
    for n in r.namelist():
        inf = r.getinfo(n)
        if cf_verbose == 1:
            show_item(inf)
        if cf_test_read and inf.is_file():
            test_read(r, inf)

    if cf_extract:
        r.extractall()
        for inf in r.infolist():
            r.extract(inf)

    if cf_test_unrar:
        r.testrar()


def test(fn, pwd):
    """Process one archive with error handling.
    """
    try:
        test_real(fn, pwd)
    except rf.NeedFirstVolume as ex:
        xprint(" --- %s is middle part of multi-vol archive (%s)---", fn, str(ex))
    except rf.Error:
        exc, msg, tb = sys.exc_info()
        xprint("\n *** %s: %s ***\n", exc.__name__, msg)
        del tb
    except IOError:
        exc, msg, tb = sys.exc_info()
        xprint("\n *** %s: %s ***\n", exc.__name__, msg)
        del tb


def main():
    """Program entry point.
    """
    global cf_verbose, cf_show_comment, cf_charset
    global cf_extract, cf_test_read, cf_test_unrar
    global cf_test_memory

    cf_backend = None
    pwd = None

    # parse args
    try:
        opts, args = getopt.getopt(sys.argv[1:], "p:C:hvcxtRMb:")
    except getopt.error as ex:
        print(str(ex), file=sys.stderr)
        sys.exit(1)

    for o, v in opts:
        if o == "-p":
            pwd = v
        elif o == "-h":
            xprint(usage)
            return
        elif o == "-v":
            cf_verbose += 1
        elif o == "-c":
            cf_show_comment = 1
        elif o == "-x":
            cf_extract = 1
        elif o == "-t":
            cf_test_read += 1
        elif o == "-T":
            cf_test_unrar = 1
        elif o == "-M":
            cf_test_memory = 1
        elif o == "-C":
            cf_charset = v
        elif o == "-b":
            cf_backend = v
        else:
            raise ValueError("unhandled switch: " + o)

    args2 = []
    for a in args:
        if a[0] == "@":
            for ln in open(a[1:], "r", encoding="utf8"):
                fn = ln[:-1]
                args2.append(fn)
        else:
            args2.append(a)
    args = args2

    if not args:
        xprint(usage)

    if cf_backend:
        cf_backend = {"7z": "sevenzip", "7zz": "sevenzip2"}.get(cf_backend, cf_backend)
        conf = {"unrar": False, "unar": False, "bsdtar": False, "sevenzip": False, "sevenzip2": False}
        assert cf_backend in conf, f"unknown backend: {cf_backend}"
        conf[cf_backend] = True
        rf.tool_setup(force=True, **conf)

    for fn in args:
        test(fn, pwd)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass




"""Read all test files.
"""

import io
from glob import glob

import pytest

import rarfile

ARCHIVE_COMMENTS = {
    "rar15-comment-lock.rar": "RARcomment -----",
    "rar15-comment.rar": "RARcomment -----",
    "rar202-comment-nopsw.rar": "RARcomment",
    "rar202-comment-psw.rar": "RARcomment",
    "rar3-comment-hpsw.rar": "RARcomment\n",
    "rar3-comment-plain.rar": "RARcomment\n",
    "rar3-comment-psw.rar": "RARcomment\n",
    "rar5-blake.rar": "RAR5 archive - blake\n",
    "rar5-crc.rar": "RAR5 archive - crc\n",
    "rar5-crc.sfx": "RAR5 archive - crc\n",
    "rar5-hpsw.rar": "RAR5 archive - hdr-password\n",
    "rar5-psw-blake.rar": "RAR5 archive - nohdr-password-blake\n",
    "rar5-psw.rar": "RAR5 archive - nohdr-password\n",
}

ARCHIVE_FILES = [
    f.replace("\\", "/")
    for f in sorted(glob("test/files/*.rar"))
    if "hpsw" not in f
    and "unicode.rar" not in f
]


def run_reading_normal(fn, comment):
    try:
        rf = rarfile.RarFile(fn)
    except rarfile.NeedFirstVolume:
        return
    if rf.needs_password():
        rf.setpassword("password")
    assert rf.strerror() is None
    assert rf.comment == comment
    for ifn in rf.namelist():
        if ifn.endswith("/"):
            continue

        info = rf.getinfo(ifn)
        if info.is_dir():
            continue
        if info.is_symlink():
            continue

        # full read
        rf.read(ifn)

        # read from stream
        item = rf.getinfo(ifn)
        f = rf.open(ifn)
        total = 0
        while True:
            buf = f.read(1024)
            if not buf:
                break
            total += len(buf)
        f.close()
        assert total == item.file_size, ifn

        # read from stream with readinto
        bbuf = bytearray(1024)
        with rf.open(ifn) as f:
            res = f.readinto(memoryview(bbuf))
            if res == 0:
                break


def run_reading_inmem(fn, comment):
    try:
        rf = rarfile.RarFile(fn)
    except rarfile.NeedFirstVolume:
        return
    if len(rf.volumelist()) > 1:
        return

    with io.open(fn, "rb") as f:
        buf = f.read()
    run_reading_normal(io.BytesIO(buf), comment)


def run_reading(fn):
    basename = fn.split("/")[-1]
    comment = ARCHIVE_COMMENTS.get(basename)
    run_reading_normal(fn, comment)
    run_reading_inmem(fn, comment)


@pytest.mark.parametrize("fn", ARCHIVE_FILES)
def test_reading(fn):
    run_reading(fn)


@pytest.mark.skipif(not rarfile._have_crypto, reason="No crypto")
def test_reading_rar3_hpsw():
    run_reading("test/files/rar3-comment-hpsw.rar")


@pytest.mark.skipif(rarfile._have_crypto, reason="Has crypto")
def test_reading_rar3_hpsw_nocrypto():
    with pytest.raises(rarfile.NoCrypto):
        run_reading("test/files/rar3-comment-hpsw.rar")


@pytest.mark.skipif(not rarfile._have_crypto, reason="No crypto")
def test_reading_rar5_hpsw():
    run_reading("test/files/rar5-hpsw.rar")


@pytest.mark.skipif(rarfile._have_crypto, reason="Has crypto")
def test_reading_rar5_hpsw_nocrypto():
    with pytest.raises(rarfile.NoCrypto):
        run_reading("test/files/rar5-hpsw.rar")


def test_reading_rar3_sfx():
    assert rarfile.is_rarfile("test/files/rar3-seektest.sfx") is False
    assert rarfile.is_rarfile_sfx("test/files/rar3-seektest.sfx") is True
    run_reading("test/files/rar3-seektest.sfx")
    run_reading("test/files/rar3-seektest.sfx")


def test_reading_rar5_crc_sfx():
    assert rarfile.is_rarfile("test/files/rar5-crc.sfx") is False
    assert rarfile.is_rarfile_sfx("test/files/rar5-crc.sfx") is True
    run_reading("test/files/rar5-crc.sfx")





# pylint: disable=comparison-with-itself,unneeded-not

from datetime import datetime, timedelta, timezone

import rarfile


def test_load_vint():
    assert rarfile.load_vint(b"\x00", 0) == (0, 1)
    assert rarfile.load_vint(b"\x80\x01", 0) == (1 << 7, 2)
    assert rarfile.load_vint(b"\x80\x80\x01", 0) == (1 << 14, 3)
    assert rarfile.load_vint(b"\x80\x80\x80\x01", 0) == (1 << 21, 4)
    assert rarfile.load_vint(b"\x80\x80\x80\x80\x01", 0) == (1 << 28, 5)
    assert rarfile.load_vint(b"\x80\x80\x80\x80\x80\x01", 0) == (1 << 35, 6)
    assert rarfile.load_vint(b"\x80" * 10 + b"\x01", 0) == (1 << 70, 11)


def test_to_datetime():
    assert rarfile.to_datetime((2020, 0, 0, 0, 0, 0)) == datetime(2020, 1, 1, 0, 0, 0)
    assert rarfile.to_datetime((2020, 60, 60, 60, 60, 60)) == datetime(2020, 12, 31, 23, 59, 59)
    assert rarfile.to_datetime((2020, 2, 30, 60, 60, 60)) == datetime(2020, 2, 28, 23, 59, 59)
    assert rarfile.to_datetime((2021, 2, 30, 60, 60, 60)) == datetime(2021, 2, 28, 23, 59, 59)


def test_to_nsdatetime():
    base = datetime(2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc)
    assert rarfile.to_nsdatetime(base, 0) is base

    res = rarfile.to_nsdatetime(base, 1000)
    assert res == base.replace(microsecond=1)
    assert isinstance(res, datetime)
    assert res.isoformat(" ") == "2020-01-01 00:00:00.000001+00:00"

    res = rarfile.to_nsdatetime(base, 1001)
    assert isinstance(res, datetime)
    assert isinstance(res, rarfile.nsdatetime)
    assert res.microsecond == 1
    assert res.nanosecond == 1001
    assert res.isoformat(" ") == "2020-01-01 00:00:00.000001001+00:00"
    assert res.isoformat(" ", "auto") == "2020-01-01 00:00:00.000001001+00:00"
    assert res.isoformat(" ", "microseconds") == "2020-01-01 00:00:00.000001+00:00"


def test_nsdatetime_cmp():
    nsdatetime = rarfile.nsdatetime

    n1 = nsdatetime(2000, 1, 1, 9, 15, 30, nanosecond=100200300, tzinfo=timezone.utc)
    n2 = nsdatetime(2000, 1, 1, 9, 15, 30, nanosecond=100200301, tzinfo=timezone.utc)
    n3 = nsdatetime(2000, 1, 1, 9, 15, 30, nanosecond=100200402, tzinfo=timezone.utc)

    d1 = datetime(2000, 1, 1, 9, 15, 30, 100100, timezone.utc)
    d2 = datetime(2000, 1, 1, 9, 15, 30, 100200, timezone.utc)
    d3 = datetime(2000, 1, 1, 9, 15, 30, 100300, timezone.utc)

    n2x = n2 + timedelta(seconds=0)
    assert not isinstance(n2x, nsdatetime)
    assert not hasattr(n2x, "_nanoseconds")
    assert n2x == d2
    assert hash(n2x) == hash(d2)
    assert hash(n2) != hash(d2)

    # compare nsdatetime only
    n1c = n1.replace()
    assert n1 == n1
    assert n1 == n1c
    assert n1 <= n1c
    assert n1 >= n1c
    assert n1 < n2
    assert n1 <= n2
    assert n1 != n2
    assert not n1 == n2
    assert n2 > n1
    assert n2 >= n1
    assert not n2 < n1
    assert not n1 > n2

    # mixed eq
    assert not d2 == n2
    assert not n2 == d2
    assert d2 != n2
    assert n2 != d2

    # mixed gt
    assert n2 > d2
    assert d3 > n2
    assert not d2 > n3
    assert not n2 > d3

    # mixed lt
    assert d1 < n2 < d3


def test_nsdatetime_astimezone():
    nsdatetime = rarfile.nsdatetime
    X1 = timezone(timedelta(hours=1), "X1")

    n1 = nsdatetime(2000, 1, 1, 9, 15, 30, nanosecond=100200402, tzinfo=timezone.utc)
    n2 = n1.astimezone(X1)
    assert n2.nanosecond == n1.nanosecond
    assert (n1.year, n1.month, n1.day) == (n2.year, n2.month, n2.day)
    assert (n1.hour, n1.minute, n1.second) == (n2.hour - 1, n2.minute, n2.second)




"""Setup script for rarfile.
"""

import re

from setuptools import setup

vrx = r"""^__version__ *= *['"]([^'"]+)['"]"""
src = open("rarfile.py").read()
ver = re.search(vrx, src, re.M).group(1)

ldesc = open("README.rst").read().strip()
sdesc = ldesc.split('\n')[0].split(' - ')[1].strip()

setup(
    name="rarfile",
    version=ver,
    description=sdesc,
    long_description=ldesc,
    author="Marko Kreen",
    license="ISC",
    author_email="markokr@gmail.com",
    url="https://github.com/markokr/rarfile",
    py_modules=['rarfile'],
    keywords=['rar', 'unrar', 'archive'],
    python_requires=">=3.10",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Archiving :: Compression",
    ]
)




"""Extract tests.
"""

import io
import os
import sys
from datetime import datetime

import pytest

import rarfile


def get_props(rf, name):
    inf = rf.getinfo(name)
    return "".join([
        inf.is_file() and "F" or "-",
        inf.is_dir() and "D" or "-",
        inf.is_symlink() and "L" or "-",
    ])


def san_unix(fn):
    return rarfile.sanitize_filename(fn, "/", False)


def san_win32(fn):
    return rarfile.sanitize_filename(fn, "/", True)


def test_sanitize_unix():
    assert san_unix("asd/asd") == "asd/asd"
    assert san_unix("asd/../asd") == "asd/asd"
    assert san_unix("c:/a/x") == r"c:/a/x"
    assert san_unix("z/./../a /b./c") == "z/a /b./c"
    assert san_unix("z<>*?:") == "z____:"


def test_sanitize_win32():
    assert san_win32("asd/asd") == "asd/asd"
    assert san_win32("asd/../asd") == "asd/asd"
    assert san_win32("c:/a/x") == "a/x"
    assert san_win32("z/./../a /b./c") == "z/a_/b_/c"
    assert san_win32("z<>*?:\\^") == "z_______"


def checktime(fn, exp_mtime):
    # cannot check subsecond precision as filesystem may not support it
    cut = len("0000-00-00 00:00:00")
    st = os.stat(fn)
    got_mtime = datetime.fromtimestamp(st.st_mtime, exp_mtime.tzinfo)
    exp_stamp = exp_mtime.isoformat(" ", "seconds")[:cut]
    got_stamp = got_mtime.isoformat(" ", "seconds")[:cut]
    assert exp_stamp == got_stamp


def checkfile(fn, data, mtime):
    with open(fn, "r", encoding="utf8") as f:
        got = f.read()
        assert got.strip() == data

    checktime(fn, mtime)


def check_subdir(rf, tmp_path):
    # pre-mkdir
    ext1 = tmp_path / "ext1"
    inf = rf.getinfo("sub/dir1/file1.txt")
    os.mkdir(ext1)
    rf.extract(inf, ext1)
    assert sorted(os.listdir(tmp_path)) == ["ext1"]
    assert os.listdir(ext1 / "sub") == ["dir1"]
    checkfile(ext1 / "sub/dir1/file1.txt", "file1", inf.mtime)

    # no mkdir
    ext2 = tmp_path / "ext2"
    inf = rf.getinfo("sub/dir2/file2.txt")
    rf.extract("sub/dir2/file2.txt", ext2)
    assert sorted(os.listdir(tmp_path)) == ["ext1", "ext2"]
    assert os.listdir(ext2 / "sub") == ["dir2"]
    checkfile(ext2 / "sub/dir2/file2.txt", "file2", inf.mtime)

    # spaced
    ext3 = tmp_path / "ext3"
    inf = rf.getinfo("sub/with space/long fn.txt")
    rf.extract("sub/with space/long fn.txt", ext3)
    checkfile(ext3 / "sub/with space/long fn.txt", "long fn", inf.mtime)

    # unicode
    ext4 = tmp_path / "ext4"
    inf = rf.getinfo("sub/üȵĩöḋè/file.txt")
    rf.extract("sub/üȵĩöḋè/file.txt", ext4)
    checkfile(ext4 / "sub/üȵĩöḋè/file.txt", "file", inf.mtime)

    # dir only
    ext5 = tmp_path / "ext5"
    inf = rf.getinfo("sub/dir2")
    rf.extract("sub/dir2", ext5)
    assert os.listdir(ext5 / "sub") == ["dir2"]
    assert os.listdir(ext5 / "sub/dir2") == []
    checktime(ext5 / "sub/dir2", inf.mtime)

    # cwd
    ext6 = tmp_path / "ext6"
    os.mkdir(ext6)
    old = os.getcwd()
    try:
        os.chdir(ext6)
        rf.extract("sub/dir1")
        assert os.listdir(".") == ["sub"]
        assert os.listdir("sub") == ["dir1"]
        assert os.listdir("sub/dir1") == []
    finally:
        os.chdir(old)

    # errors
    with pytest.raises(io.UnsupportedOperation):
        rf.open("sub/dir1")


@pytest.mark.parametrize("fn", [
    "test/files/rar3-subdirs.rar",
    "test/files/rar5-subdirs.rar",
])
def test_subdirs(fn, tmp_path):
    with rarfile.RarFile(fn) as rf:
        check_subdir(rf, tmp_path)


@pytest.mark.parametrize("fn", [
    "test/files/rar3-readonly-unix.rar",
    "test/files/rar3-readonly-win.rar",
    "test/files/rar5-readonly-unix.rar",
    "test/files/rar5-readonly-win.rar",
])
def test_readonly(fn, tmp_path):
    with rarfile.RarFile(fn) as rf:
        assert get_props(rf, "ro_dir") == "-D-"
        assert get_props(rf, "ro_dir/ro_file.txt") == "F--"

        rf.extractall(tmp_path)

    assert os.access(tmp_path / "ro_dir/ro_file.txt", os.R_OK)
    assert not os.access(tmp_path / "ro_dir/ro_file.txt", os.W_OK)

    if sys.platform != "win32":
        assert os.access(tmp_path / "ro_dir", os.R_OK)
        assert not os.access(tmp_path / "ro_dir", os.W_OK)


@pytest.mark.parametrize("fn", [
    "test/files/rar3-symlink-unix.rar",
    "test/files/rar5-symlink-unix.rar",
])
def test_symlink(fn, tmp_path):
    with rarfile.RarFile(fn) as rf:
        assert get_props(rf, "data.txt") == "F--"
        assert get_props(rf, "data_link") == "--L"
        assert get_props(rf, "random_link") == "--L"

        rf.extract('data.txt', tmp_path)
        rf.extract('data_link', tmp_path)
        with pytest.raises(rarfile.BadSymLinkError):
            rf.extract('random_link', tmp_path)

        assert sorted(os.listdir(tmp_path)) == ["data.txt", "data_link"]

        data = rf.getinfo("data.txt")
        data_link = rf.getinfo("data_link")

        assert not data.is_symlink()
        assert data_link.is_symlink()

        assert rf.read(data) == b"data\n"
        assert rf.read(data_link) == b"data.txt"

        assert os.path.isfile(tmp_path / "data.txt")
        assert os.path.islink(tmp_path / "data_link")

        # str - work around pypy3 bug
        assert os.readlink(str(tmp_path / "data_link")) == "data.txt"


def test_symlink_win(tmp_path):
    fn = "test/files/rar5-symlink-win.rar"
    with rarfile.RarFile(fn) as rf:
        assert get_props(rf, "content/dir1") == "-D-"
        assert get_props(rf, "content/dir2") == "-D-"
        assert get_props(rf, "content/file.txt") == "F--"
        assert get_props(rf, "links/bad_link") == "--L"
        assert get_props(rf, "links/dir_junction") == "--L"
        assert get_props(rf, "links/dir_link") == "--L"
        assert get_props(rf, "links/file_link") == "--L"

        with pytest.warns(rarfile.UnsupportedWarning):
            rf.extract('content/dir1', tmp_path)
            rf.extract('content/dir2', tmp_path)
            rf.extract('content/file.txt', tmp_path)
            rf.extract('links/dir_junction', tmp_path)
            rf.extract('links/dir_link', tmp_path)
            rf.extract('links/file_link', tmp_path)

        with pytest.raises(rarfile.BadSymLinkError):
            rf.extract('links/bad_link', tmp_path)

        assert sorted(os.listdir(tmp_path)) == ["content", "links"]
        assert sorted(os.listdir(tmp_path / "content")) == ["dir1", "dir2", "file.txt"]
        assert sorted(os.listdir(tmp_path / "links")) == ["dir_link", "file_link"]

        assert os.path.islink(tmp_path / "links/dir_link")
        assert os.path.islink(tmp_path / "links/file_link")

@pytest.mark.parametrize("fn", [
    "test/files/rar3-old.rar",
    "test/files/rar3-vols.part1.rar",
    "test/files/rar5-vols.part1.rar",
])
def test_vols(fn, tmp_path):
    with rarfile.RarFile(fn) as rf:
        rarfile.FORCE_TOOL = True
        try:
            rf.extractall(str(tmp_path))
        finally:
            rarfile.FORCE_TOOL = False


        assert sorted(os.listdir(tmp_path)) == ["vols"]
        assert sorted(os.listdir(tmp_path / "vols")) == ["bigfile.txt", "smallfile.txt"]

        assert os.path.isfile(tmp_path / "vols" / "bigfile.txt")
        assert os.path.isfile(tmp_path / "vols" / "smallfile.txt")



@pytest.mark.skipif(sys.platform == "win32", reason="symlink semantics differ on Windows")
def test_symlink_chained_traversal_blocked(tmp_path):
    """Regression test for symlink-chained zip-slip (CWE-22 + CWE-59).

    The fixture rar5-evil-symlink-traversal.rar contains two members:
      1. up           -- symlink whose target is the relative path '..'
      2. up/pwned.txt -- regular file whose extraction path traverses
                         the symlink above and lands outside the extract dir.

    Without the realpath containment check in _extract_one, extractall
    silently writes pwned.txt outside the extraction directory (in
    tmp_path rather than tmp_path/extract).
    """
    extract = tmp_path / "extract"
    extract.mkdir()

    fn = "test/files/rar5-evil-symlink-traversal.rar"
    with rarfile.RarFile(fn) as rf:
        with pytest.raises(rarfile.BadSymLinkError):
            rf.extractall(str(extract))

    # test if pre-existing symlink is detected
    up = extract / "up"
    up.symlink_to('..')
    assert str(up.readlink()) == '..'

    with pytest.raises(rarfile.BadRarFile):
        rf.extract('up/pwned.txt', extract)

    # The second member would land at tmp_path/pwned.txt without the guard.
    assert not (tmp_path / "pwned.txt").exists()




"""Alt tool tests
"""

import sys
import os

import pytest

import rarfile


def have_tool(name):
    for dn in os.get_exec_path():
        if os.path.isfile(os.path.join(dn, name)):
            return True
        if os.path.isfile(os.path.join(dn, name + ".exe")):
            return True
    return False


def tool_setup(unrar=False, unar=False, bsdtar=False, sevenzip=False, sevenzip2=False):
    rarfile.FORCE_TOOL = True
    rarfile.tool_setup(unrar=unrar, unar=unar, bsdtar=bsdtar,
                       sevenzip=sevenzip, sevenzip2=sevenzip2,
                       force=True)


def install_unrar_tool():
    tool_setup(unrar=True)


def install_unar_tool():
    tool_setup(unar=True)


def install_bsdtar_tool():
    tool_setup(bsdtar=True)


def install_7z_tool():
    tool_setup(sevenzip=True)


def install_7zz_tool():
    tool_setup(sevenzip2=True)


def uninstall_alt_tool():
    rarfile.FORCE_TOOL = False
    rarfile.tool_setup(force=True)


def test_read_rar3():
    with rarfile.RarFile("test/files/seektest.rar") as rf:
        for fn in rf.namelist():
            rf.read(fn)


def test_read_vols():
    with rarfile.RarFile("test/files/rar3-old.rar") as rf:
        for fn in rf.namelist():
            rf.read(fn) # old
    with rarfile.RarFile("test/files/rar3-vols.part1.rar") as rf:
        for fn in rf.namelist():
            rf.read(fn) # rar3-new
    with rarfile.RarFile("test/files/rar5-vols.part1.rar") as rf:
        for fn in rf.namelist():
            rf.read(fn) # rar5


def test_unrar_tool():
    install_unrar_tool()
    try:
        test_read_rar3()
        test_read_vols()

        with rarfile.RarFile("test/files/rar3-comment-plain.rar") as rf:
            rf.read("file1.txt")
            rf.read("file2.txt")

        with rarfile.RarFile("test/files/rar3-comment-psw.rar") as rf:
            rf.setpassword("password")
            rf.read("file1.txt")
    finally:
        uninstall_alt_tool()


@pytest.mark.skipif(sys.platform == "win32", reason="unar not available on Windows")
@pytest.mark.skipif(not have_tool(rarfile.UNAR_TOOL), reason="unar not installed")
def test_unar_tool():
    install_unar_tool()
    try:
        test_read_rar3()
        test_read_vols()

        with rarfile.RarFile("test/files/rar3-comment-plain.rar") as rf:
            rf.read("file1.txt")
            rf.read("file2.txt")

        with rarfile.RarFile("test/files/rar3-comment-psw.rar") as rf:
            rf.setpassword("password")
            rf.read("file1.txt")
    finally:
        uninstall_alt_tool()


@pytest.mark.skipif(not have_tool(rarfile.BSDTAR_TOOL), reason="bsdtar not installed")
def test_bsdtar_tool():
    install_bsdtar_tool()
    try:
        #test_read_rar3()
        #test_read_vols()

        with rarfile.RarFile("test/files/rar3-comment-plain.rar") as rf:
            rf.read("file1.txt")
            rf.read("file2.txt")

        with pytest.raises(rarfile.RarCannotExec):
            with rarfile.RarFile("test/files/rar3-comment-psw.rar") as rf:
                rf.setpassword("password")
                rf.read("file1.txt")
    finally:
        uninstall_alt_tool()


@pytest.mark.skipif(not have_tool(rarfile.SEVENZIP_TOOL), reason="7z not installed")
def test_7z_tool():
    install_7z_tool()
    try:
        #test_read_rar3()
        test_read_vols()

        with rarfile.RarFile("test/files/rar3-comment-plain.rar") as rf:
            rf.read("file1.txt")
            rf.read("file2.txt")

        with rarfile.RarFile("test/files/rar3-comment-psw.rar") as rf:
            rf.setpassword("password")
            rf.read("file1.txt")
    finally:
        uninstall_alt_tool()


@pytest.mark.skipif(not have_tool(rarfile.SEVENZIP2_TOOL), reason="7zz not installed")
def test_7zz_tool():
    install_7zz_tool()
    try:
        #test_read_rar3()
        test_read_vols()

        with rarfile.RarFile("test/files/rar3-comment-plain.rar") as rf:
            rf.read("file1.txt")
            rf.read("file2.txt")

        with rarfile.RarFile("test/files/rar3-comment-psw.rar") as rf:
            rf.setpassword("password")
            rf.read("file1.txt")
    finally:
        uninstall_alt_tool()


# test popen errors

def test_popen_fail():
    with pytest.raises(rarfile.RarCannotExec):
        rarfile.custom_popen(["missing-unrar-exe"])

    if sys.platform != "win32":
        with pytest.raises(rarfile.RarCannotExec):
            rarfile.custom_popen(["./test/files/rar5-blake.rar.exp"])


def test_check_returncode():
    errmap = rarfile.UNRAR_CONFIG["errmap"]

    assert not rarfile.check_returncode(0, "", errmap)

    with pytest.raises(rarfile.RarFatalError):
        rarfile.check_returncode(2, "x", errmap)
    with pytest.raises(rarfile.RarUnknownError):
        rarfile.check_returncode(100, "", errmap)
    with pytest.raises(rarfile.RarUserBreak):
        rarfile.check_returncode(255, "", errmap)
    with pytest.raises(rarfile.RarSignalExit):
        rarfile.check_returncode(-11, "", errmap)

    errmap = rarfile.UNAR_CONFIG["errmap"]
    with pytest.raises(rarfile.RarUnknownError):
        rarfile.check_returncode(2, "", errmap)


# own cli tests

def cli(*args):
    try:
        rarfile.main(args)
        return 0
    except SystemExit as ex:
        return int(ex.code)
    except Exception as ex:
        sys.stderr.write(str(ex) + "\n")
        return 1


def test_cli_list(capsys):
    assert cli("-l", "test/files/rar3-old.rar") == 0
    res = capsys.readouterr()
    assert "bigfile" in res.out


def test_cli_testrar(capsys):
    assert cli("-t", "test/files/rar3-old.rar") == 0
    res = capsys.readouterr()
    assert not res.err


def test_cli_extract(capsys, tmp_path):
    assert cli("-e", "test/files/rar3-old.rar", str(tmp_path)) == 0
    res = capsys.readouterr()
    assert not res.err


def test_cli_help(capsys):
    assert cli("--help") == 0
    res = capsys.readouterr()
    assert "option" in res.out




"""Hashing tests.
"""

import hashlib
from binascii import hexlify, unhexlify

import rarfile
from rarfile import Blake2SP, CRC32Context, NoHashContext, Rar3Sha1


def tohex(data):
    """Return hex string."""
    return hexlify(data).decode("ascii")


def test_nohash():
    assert NoHashContext("").hexdigest() is None
    assert NoHashContext("asd").hexdigest() is None
    md = NoHashContext()
    md.update("asd")
    assert md.digest() is None


def test_crc32():
    assert CRC32Context(b"").hexdigest() == "00000000"
    assert CRC32Context(b"Hello").hexdigest() == "f7d18982"
    assert CRC32Context(b"Bye").hexdigest() == "4f7ad7d4"

    md = CRC32Context()
    md.update(b"He")
    md.update(b"ll")
    md.update(b"o")
    assert md.hexdigest() == "f7d18982"


def xblake2sp(xdata):
    data = unhexlify(xdata)
    md = Blake2SP()
    md.update(data)
    return md.hexdigest()


def xblake2sp_slow(xdata):
    data = unhexlify(xdata)
    md = Blake2SP()
    buf = memoryview(data)
    pos = 0
    while pos < len(buf):
        md.update(buf[pos: pos + 3])
        pos += 3
    return md.hexdigest()


def test_blake2sp():
    assert Blake2SP(b"").hexdigest() == "dd0e891776933f43c7d032b08a917e25741f8aa9a12c12e1cac8801500f2ca4f"
    assert Blake2SP(b"Hello").hexdigest() == "0d6bae0db99f99183d060f7994bb94b45c6490b2a0a628b8b1346ebea8ec1d66"

    assert xblake2sp("") == "dd0e891776933f43c7d032b08a917e25741f8aa9a12c12e1cac8801500f2ca4f"
    assert xblake2sp("00") == "a6b9eecc25227ad788c99d3f236debc8da408849e9a5178978727a81457f7239"

    long1 = "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f202122232425262728292a2b2c2d2e2f3031"
    assert xblake2sp(long1) == "270affa6426f1a515c9b76dfc27d181fc2fd57d082a3ba2c1eef071533a6dfb7"

    long2 = long1 * 20
    assert xblake2sp(long2) == "24a78d92592d0761a3681f32935225ca55ffb8eb16b55ab9481c89c59a985ff3"
    assert xblake2sp_slow(long2) == "24a78d92592d0761a3681f32935225ca55ffb8eb16b55ab9481c89c59a985ff3"


def test_rar3_sha1():
    for n in range(0, 200):
        data = bytearray(range(n))
        h1 = hashlib.sha1(data).hexdigest()
        h2 = Rar3Sha1(data).hexdigest()
        assert h1 == h2

    data = bytearray([(i & 255) for i in range(2000)])
    x1 = hashlib.sha1()
    x2 = Rar3Sha1()
    for step in (3, 17, 67, 128, 157):
        pos = 0
        while pos < len(data):
            pos2 = pos + step
            if pos2 > len(data):
                pos2 = len(data)
            x1.update(data[pos:pos2])
            x2.update(data[pos:pos2])
            assert x1.hexdigest() == x2.hexdigest()
            pos = pos2


def test_rar3_s2k():
    exp = ("a160cb31cb262e9231c0b6fc984fbb0d", "aa54a659fb0c359b30f353a6343fb11d")
    key, iv = rarfile.rar3_s2k(b"password", unhexlify("00FF00"))
    assert (tohex(key), tohex(iv)) == exp
    key, iv = rarfile.rar3_s2k("password", unhexlify("00FF00"))
    assert (tohex(key), tohex(iv)) == exp

    exp = ("ffff33ffaf31987c899ccc2f965a8927", "bdff6873721b247afa4f978448a5aeef")
    key, iv = rarfile.rar3_s2k("p" * 28, unhexlify("1122334455667788"))
    assert (tohex(key), tohex(iv)) == exp
    exp = ("306cafde28f1ea78c9427c3ec642c0db", "173ecdf574c0bfe9e7c23bdfd96fa435")
    key, iv = rarfile.rar3_s2k("p" * 29, unhexlify("1122334455667788"))
    assert (tohex(key), tohex(iv)) == exp




"""Test seeking on files.
"""

import io

import pytest

import rarfile

ARC = "test/files/seektest.rar"

_WHENCE = 0


def do_seek(f, pos, lim, size=None):
    global _WHENCE
    ofs = pos * 4
    fsize = lim * 4

    if ofs < 0:
        exp = 0
    elif ofs > fsize:
        exp = fsize
    else:
        exp = ofs

    if size:
        cur = f.tell()
        if _WHENCE == 0:
            f.seek(ofs, _WHENCE)
        elif _WHENCE == 1:
            f.seek(ofs - cur, _WHENCE)
        else:
            assert _WHENCE == 2
            f.seek(ofs - size, _WHENCE)
        _WHENCE = (_WHENCE + 1) % 3
    else:
        f.seek(ofs)

    got = f.tell()

    assert got == exp
    ln = f.read(4)
    if got == fsize and ln:
        raise ValueError("unexpected read")
    if not ln and got < fsize:
        raise ValueError("unexpected read failure")
    if ln:
        spos = int(ln)
        assert spos * 4 == got


def run_seek(rf, fn):
    inf = rf.getinfo(fn)
    cnt = int(inf.file_size / 4)
    f = rf.open(fn)

    with pytest.raises(ValueError):
        f.seek(0, -1)
    with pytest.raises(ValueError):
        f.seek(0, 3)

    do_seek(f, int(cnt / 2), cnt)
    do_seek(f, 0, cnt)

    for i in range(int(cnt / 2)):
        do_seek(f, i * 2, cnt, inf.file_size)

    for i in range(cnt):
        do_seek(f, i * 2 - int(cnt / 2), cnt, inf.file_size)

    for i in range(cnt + 10):
        do_seek(f, cnt - i - 5, cnt, inf.file_size)

    f.close()


def run_arc(arc, desc):
    files = ["stest1.txt", "stest2.txt"]
    rf = rarfile.RarFile(arc)
    for fn in files:
        run_seek(rf, fn)


def test_seek_filename():
    run_arc(ARC, "fn")


def test_seek_bytesio():
    # filelike: io.BytesIO, io.open()
    with open(ARC, "rb") as f:
        data = f.read()
    run_arc(io.BytesIO(data), "io.BytesIO")


def test_seek_open():
    # filelike: file()
    with open(ARC, "rb") as f:
        run_arc(f, "open")


def test_seek_ioopen():
    # filelike: io.open()
    with io.open(ARC, "rb") as f:
        run_arc(f, "io.open")


def run_seek_middle(fn, entry):
    rar = rarfile.RarFile(fn)
    file = rar.open(entry)
    assert file.read(1) == b"0"
    assert file.read(1) == b"0"
    assert file.read(1) == b"0"
    assert file.read(1) == b"\n"
    file.read()

    file.seek(0)
    assert file.read(1) == b"0"
    assert file.read(1) == b"0"
    assert file.read(1) == b"0"
    assert file.read(1) == b"\n"
    file.read()

    file.seek(2)
    assert file.read(1) == b"0"
    assert file.read(1) == b"\n"
    file.read()

def test_seek_middle1():
    run_seek_middle("test/files/seektest.rar", "stest1.txt")

def test_seek_middle2():
    run_seek_middle("test/files/seektest.rar", "stest2.txt")

